/*
   Do not modify this file.
*/

import java.util.Random;

/**
   This simulates a first-in-first-out (stack) buffer.
*/
public class Buffer
{
   protected final int size;
   protected final char[] buf;
   protected int currentIndex;
   protected final char BEGINNING;
   protected final char END;
   protected final char EMPTY;
   protected final char FILL;
   protected final static Random random = new Random();

   /**
      Initialize a partially filled buffer.

      A currentIndex value of 0 creates an initially empty buffer.

      A currentIndex value of size creates an initially full buffer.

      The buffer array actually holds size + 2 characters to allow for
      one character to visually mark the beginning of the buffer and
      another character to visually mark the end of the buffer.

      @param size          the size of the buffer
      @param currentIndex  initially fill the buffer up to this index
      @param c1  the left bracket for the string representation of the buffer
      @param c2  the char for the filled part of the buffer
      @param c3  the char for the empty part of the buffer
      @param c4  the right bracket for the string representation of the buffer
      @throws IllegalArgumentException if currentIndex is outside of the buffer
   */
   Buffer(final int size, final int currentIndex,
          final char c1, final char c2, final char c3, final char c4)
   {
      // currentIndex == 0     initializes an empty buffer
      // currentIndex == size  initializes a full buffer
      if (currentIndex < 0 || currentIndex > size)
      {
         throw new IllegalArgumentException("illegal buffer initialization");
      }

      this.BEGINNING = c1;
      this.FILL = c2;
      this.EMPTY = c3;
      this.END = c4;

      this.size = size;
      this.buf = new char[size + 2];
      this.currentIndex = currentIndex;
      this.buf[0] = BEGINNING;  // Mark the beginning of the buffer.
      for (int i = 1; i <= currentIndex; ++i)
      {
         this.buf[i] = (char)(FILL + random.nextInt(10)); // Mark the full part of the buffer.
      }
      for (int i = currentIndex + 1; i <= size; ++i)
      {
         this.buf[i] = EMPTY;   // Mark the empty part of the buffer.
      }
      this.buf[size + 1] = END; // Mark the end of the buffer.
   }


   public boolean isFull()
   {
      return currentIndex == size;
   }


   public boolean isEmpty()
   {
      return currentIndex == 0;
   }


   public void put()
   throws InterruptedException // because of the subclasses
   {
      put((char)(FILL + random.nextInt(10)));
   }


   public char get()
   throws InterruptedException // because of the subclasses
   {
      return get(EMPTY);
   }


   /**
      Insert the given character at the next available buffer location.

      Every time we do an operation on this buffer, it prints out its state.

      @param c  character to put in the buffer
      @thows IllegalStateException when the buffer is full
   */
   public void put(char c)
   throws InterruptedException // because of the subclasses
   {
      if (currentIndex == size)
      {
         // This buffer is not bounded so this buffer
         // throws an exception when it tries to overflow.
         throw new IllegalStateException("buffer is full");
      }

      ++currentIndex;
      Thread.yield(); // Allow race conditions to happen more often.
      buf[currentIndex] = c;
      System.out.println(this);  // Print the state of this buffer.
      System.out.flush();
   }


   /**
      Remove from the buffer the character at the buffer's current index.
      Mark the newly emptied spot in the buffer with the given character.
      Return the character that was removed from the buffer.

      Every time we do an operation on this buffer, it prints out its state.

      @param c  character to use to mark the newly emptied position in the buffer
      @return the character removed from the buffer
      @thows IllegalStateException when the buffer is empty
   */
   public char get(char c)
   throws InterruptedException // because of the subclasses
   {
      if (currentIndex == 0)
      {
         // This buffer is not bounded so this buffer
         // throws an exception when it tries to underflow.
         throw new IllegalStateException("buffer is empty");
      }

      char result = buf[currentIndex];
      buf[currentIndex] = c;
      Thread.yield(); // Allow race conditions to happen more often.
      --currentIndex;
      System.out.println(this);  // Print the state of this buffer.
      System.out.flush();
      return result;
   }


   /**
      Create a String representation of this buffer showing the
      buffer's contents in the first line and the current position
      of the buffer's index in the second line.

      This method is synchronized so that this buffer prints itself
      out properly even when its used in a multithreaded program.
   */
   @Override
   public synchronized String toString()
   {
      String result = "";
      // first line
      for (int i = 0; i < size + 2; ++i)
      {
         result += buf[i];
      }
      result += "\n";
      // second line
      for (int i = 0; i < this.currentIndex; ++i)
      {
         result += " ";
      }
      result += "^\n";
      return result;
   }
}
